--[[
	Lighting library by Brent "Xan the Dragon" D.
	Revision 2
	
	Note:
		Revision 1 was not released to the public as it was very laggy and used a terrible method
		involving iterating through all pixels on the screen (by a specific, specified grid unit)
		and finding if a certain point was within the radius of a light. Don't be a Xan.
		
	Revision 2 uses a MUCH MUCH MUCH faster method which uses shaders.
	
	Limitations:
		Only up to 128 lights can be registered at one time. This is a limit of GLSL and how I have my arrays set up.
		As of now, lighting ignores any obstructions. This will be changed soon.
	
	API
		Properties:
			Brightness						A value from 0 to 1 where 0 = pitch black with no lights and 1 = fully bright no matter what
			
		Methods:
			register(Light)					Register a light object, adding it to the draw list. (See Light.lua)
			unregister(Light)				Unregister a light object, removing it from the draw list.
			activate()						Activates the shader. Anything drawn while Lighting is active will be affected by the shader.
			deactivate()					Deactivates the shader, returning all drawing to its normal style.
			update()						Tells the lighting engine to update. ANY TIME YOU SET A LIGHT'S ENABLED STATE, FIRE THIS! (Alternatively, read Light.lua's documentation!)
	
--]]

local LightCount = 0
local Brightness = 0.075 --0 = black, 1 = no lighting
local reg_Shader = love.graphics.newShader("lightlib/shader.glsl") --If this line is erroring, try fixing it. Most Helpful Advice 2017. All joking aside, just make it point to the shader.glsl file relative to the same directory as Main.lua
reg_Shader:send("ambient_color", {Brightness * 255, Brightness * 255, Brightness * 255, 255})

local Lighting = {}
local Lights = {}
local Meta = {}

local cPos = {} --All light positions
local cCol = {} --All light colors
local cRad = {} --All light radius.. radiuses? radii?
local cStx = {} --All light states
local cRef = {} --All light references. The indices in this table are light objects, and their values are the index of the light's value in the 4 tables above

setmetatable(Lighting, Meta)

local function register(light)
	if not light then return end
	Lights[light] = light; --This is weird but nice to have
	cPos = {}
	cCol = {}
	cRad = {}
	cStx = {}
	cRef = {}
	for Index, v in pairs(Lights) do
		table.insert(cPos, v.Position)
		table.insert(cCol, v.Color)
		table.insert(cRad, v.Radius)
		table.insert(cStx, v.Enabled) --Fun fact: This is how you do a ternary operator in lua. The catch is that the thing between "and" and "or" must not be false or nil.
		cRef[v] = #cPos --I cound do the length of any of the above.
	end
	LightCount = LightCount + 1
	reg_Shader:send("light_pos", unpack(cPos))
	reg_Shader:send("light_colors", unpack(cCol))
	reg_Shader:send("light_radius", unpack(cRad))
	reg_Shader:send("light_state", unpack(cStx))
	reg_Shader:send("light_count", LightCount)
end

local function unregister(light)
	if not light then return end
	Lights[light] = nil;
	cPos = {}
	cCol = {}
	cRad = {}
	cStx = {}
	cRef = {}
	for Index, v in pairs(Lights) do
		table.insert(cPos, v.Position)
		table.insert(cCol, v.Color)
		table.insert(cRad, v.Radius)
		table.insert(cStx, v.Enabled)
		cRef[v] = #cPos --I cound do the length of any of the above.
	end
	LightCount = LightCount - 1
	reg_Shader:send("light_pos", unpack(cPos))
	reg_Shader:send("light_colors", unpack(cCol))
	reg_Shader:send("light_radius", unpack(cRad))
	reg_Shader:send("light_state", unpack(cStx))
	reg_Shader:send("light_count", LightCount)
end

local function activate()
	love.graphics.setShader(reg_Shader)
end

local function deactivate()
	love.graphics.setShader()
end

local function update(light)
	if light then
		local idx = cRef[light]
		if idx then
			cPos[idx] = light.Position
			cCol[idx] = light.Color
			cRad[idx] = light.Radius
			cStx[idx] = light.Enabled
			reg_Shader:send("light_pos", unpack(cPos))
			reg_Shader:send("light_colors", unpack(cCol))
			reg_Shader:send("light_radius", unpack(cRad))
			reg_Shader:send("light_state", unpack(cStx))
			--If the index for the light was nil, it means we set some property (and the updater) of a light before registering it.
			--In this instance we will do nothing.
		end
	else
		cPos = {}
		cCol = {}
		cRad = {}
		cStx = {}
		cRef = {}
		LightCount = 0
		for Index, v in pairs(Lights) do
			table.insert(cPos, v.Position)
			table.insert(cCol, v.Color)
			table.insert(cRad, v.Radius)
			table.insert(cStx, v.Enabled)
			cRef[v] = #cPos --I cound do the length of any of the above.
			LightCount = LightCount + 1
		end
		reg_Shader:send("light_pos", unpack(cPos))
		reg_Shader:send("light_colors", unpack(cCol))
		reg_Shader:send("light_radius", unpack(cRad))
		reg_Shader:send("light_state", unpack(cStx))
		reg_Shader:send("light_count", LightCount)
	end
end

Meta.__index = function (Table, Index)
	if Table == Lighting then
		if Index == "register" then
			return register
		elseif Index == "unregister" then
			return unregister
		elseif Index == "activate" then
			return activate
		elseif Index == "deactivate" then
			return deactivate
		elseif Index == "update" then
			return update
		elseif Index == "Brightness" then
			return Brightness
		end
	end
end

Meta.__newindex = function (Table, Index, Value)
	if Table == Lighting then
		if Index == "Brightness" then
			Brightness = Value
			local v = Brightness * 255;
			reg_Shader:send("ambient_color", {v, v, v, 255})
		end
	end
end

return Lighting